package WGP::Sequence::ApproxMatch;

use 5.008005;
use strict;
use warnings;
use Carp qw(cluck confess croak);

our $VERSION = '0.01';


sub new {

  my $class = shift;
  ( bless {}, $class )->_init( {@_} );
  
}

sub _init {

    my ($self, $args) = @_;
#    $self->{pmatch}   = $args->{pmatch}   || 0;      #binary executable, path and name, i.e. /bin/pmatch
#    $self->{distance} = $args->{distance} || 0;                                 #num mismatch/indels... 
#    $self->{target}   = $args->{target}   || confess "target file not defined"; #fasta for pmatch target
    $self->{query} = $args->{query} || confess "query file not defined";  #fasta
    $self->{min_perc_qcoverage} = $args->{min_perc_qcoverage}  || confess "minimum required percentage query coverage not defined";
    $self->{dc_template} = $args->{dc_template} || 0;  #dc_template
    $self->{dc_target} = $args->{dc_target} || 0;
    $self->{dc_machine} = $args->{dc_machine} || 0;
    $self->{out} = $args->{out} || 0;                                 #output file, write to STDOUT if zero
    return $self;
}


#sub run_primermatch {
#
#    my $self = shift;
#    confess "no query file"  unless ( -e $self->{query}  );
#    confess "no target file" unless ( -e $self->{target} );
#    confess "no pmatch binary" unless ( -e $self->{pmatch} );
##    %P  The FASTA header (defline) of the primer,
##    %h  FASTA header (defline) of the sequence entry containing the alignment.
##    %s  Start position of the alignment within the FASTA entry (space based).
##    %e  End position of the alignment in the FASTA entry (space based).
##    %D  Length difference between primer and aligned sequence
##    %d  edit distance
#
#    my $cmd = $self->{pmatch} 
#              . " -i " . $self->{target} 
#              . " -F " . $self->{query} 
#              . " -k " . $self->{distance}
#              . " -A " . "\"\%P,\%h,\%s,\%e,\%l,\%d\\n\"";
#
#    $cmd   .=   " -o " . $self->{out} if ($self->{out});
#    my $exitStat = system($cmd); 
#    confess "pmatch died with exit status: $exitStat" if ($exitStat);
#
#}

sub run_decypher {

    my $self = shift;

    confess "no template file" unless ( -e $self->{dc_template}  );
    confess "no query file" unless ( -e $self->{query} );
    confess "no dc_target"  unless ( $self->{dc_target} );
    confess "no dc machine"  unless ( $self->{dc_machine} );
    
    my $targstr = join (" -targ ", split (',', $self->{dc_target}) );

    my $cmd =   "dc_template "
              . " -template " . $self->{dc_template}
              . " -query "    . $self->{query}
              . " -targ "     . $targstr
              . " -mach "     . $self->{dc_machine};

    my $dc_socString = `$cmd`;  #uns the dc job

    my ($dc_jobId) = $dc_socString =~ /(SOC.+)/;
    confess "Error: dc job id could not retrived upon execution of dc_template command. Got string: $dc_socString" unless ($dc_jobId);
    my $dc_deliverCmd =   "dc_deliver "
                        . $dc_jobId 
                        . " -mach " . $self->{dc_machine};
    $dc_deliverCmd .=  " > " .  $self->{out} if ($self->{out});
    my $exitStat = system($dc_deliverCmd);
    confess "Error: dc_deliver exited with errors, exit stat: $exitStat" if ($exitStat);
    $self->_fix_dc_deliver_bug($self->{out});  #duplicate output error
}

sub _fix_dc_deliver_bug {

    my $self = shift;
    my $dc_deliver_outfile = shift;
    open (FH, $dc_deliver_outfile) or confess $!;
    my (%line, $header);
    while (<FH>) {
        $header = $_ if (/ALIGNMENTLENGTH/);
        $line{$_} = 1;  #erases duplicates implicitly
    }
    close FH;
    unlink $dc_deliver_outfile;
    open (OUT, ">$dc_deliver_outfile") or confess $!;
    print OUT $header;
    for my $key (keys %line ) {
        next if ($key =~ /ALIGNMENTLENGTH/);
        my @args = split (/\t/, $key);
        next if ($args[2] < $self->{min_perc_qcoverage});
        print OUT $key;
    }
    close OUT;

}

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

WGP::Sequence::ApproxMatch -

=head1 SYNOPSIS

  use WGP::Sequence::ApproxMatch;

=head1 DESCRIPTION

Stub documentation for WGP::Sequence::ApproxMatch

=head2 EXPORT

None by default.



=head1 SEE ALSO

=head1 AUTHOR

F. Yalcin E<lt>fy@keygene.comE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009 by KeyGene NV

=cut
