#!/usr/bin/env perl

# Use some fancy forking so that we can apply a series of independent
# filters to massage an FPC band file
# Example File. (omit leading '#')
#E00BAC3084 3 BAC_WGP
#5698
#6782
#15494
#-1
#E00BAC8288 1 BAC_WGP
#8538
#-1
#E00BAC2468 3 BAC_WGP
#14674
#15061
#15186
#-1


use strict;
use warnings;
use FileHandle;
use English;
use Data::Dumper;
use Carp;
#use Math::Random;


my $maxtagid = 65535;

sub FakeData {
    my ($numchroms, $numbacs, $numtags) = @_;


    my ($readfh, $writefh) = FileHandle::pipe;


    foreach my $bac (1..$numbacs) {
	my $chrom = int(rand($numchroms)+1);
	my $size = rand(100_000)+20_000;
    }



    print $writefh "31.E00BAC3084 5 BAC_WGP
17094
17744
20614
51364
52009
-1
31.E00BAC8288 1 BAC_WGP
25614
-1
31.E00BAC8146 16 BAC_WGP
34599
36514
36879
36914
37809
38044
38759
49349
49714
51644
52449
52724
53024
59974
59989
64884
-1
31.E00BAC5236 18 BAC_WGP
12664
12929
13029
13729
14074
14249
14704
14719
15454
16334
16589
16939
16999
17029
18244
18499
18634
21369
-1
31.E00BAC974 16 BAC_WGP
12214
12424
12784
13069
14429
14474
15204
15459
15559
16509
16689
17459
18054
18869
19724
20294
-1
";
    return $readfh;
} 

# mainly used during testing, but does allow you to keep only the top N records
sub head {
    my $lines = shift || 20;

    my $pid;
    return if $pid = open(STDOUT, "|-");
    die "cannot fork: $!" unless defined $pid;

    while (<STDIN>) {
        print;
        last unless --$lines ;
    } 
    exit;
} 

sub tagFilter {
    my ($forbiddenFns, $translateFn, $badbacs) = @_;



    my $pid;
    return if $pid = open(STDOUT, "|-");
    die "cannot fork: $!" unless defined $pid;

    my $forbidden = {};

    { 
	
	local $INPUT_RECORD_SEPARATOR = "\n";

	#my $forbfh = FileHandle->new($forbiddenTagsFn);
	my $translateFh = FileHandle->new($translateFn) or die("$!:$translateFn");
	my $translate = {};
	while ( my $line = <$translateFh>) {
	    my ($seq, $id, $bacstr) = $line =~ /^(\S+)\s+(\d+)\s+(.*)/;
	    
	    next unless $seq;
	    next unless $id;



	    foreach my $str ( 
			      'TTACTCGTATGC',
			      'TTACTCGTATG',
			      'TTACTCGTAT',
			      'TTACTCGTA',
			      'TTACTCGT',
			      'TTACTCG',
			      'TTACTC',
			      'TTACT',
			      'TTAC',
			      ) {
		if ($seq =~ /$str$/) {
		    #print "$seq match MseI. leaving it in, but trimming for .seq file\n";
		    #$forbidden->{$id} = 1 ;
		    $seq =~ s/$str$/TTA/;
		}
	    }


	    $translate->{$seq} = $id;


	    my $bactagcount = $bacstr =~ tr/,/,/;
	    if ($bactagcount < 2) {
		$forbidden->{$id} = 1 ;
		print STDERR "Filtering\t$id\tsingleton tag\n";
	    }


	}


	my $minrun = 5;

	foreach my $seq (keys %$translate) {
	    #print "consider $seq\n";
	    my $trans = $translate->{$seq};
	    my $homofail = 0;
	    $homofail = 1 if $seq =~ /a{$minrun}/i;
	    $homofail = 1 if $seq =~ /t{$minrun}/i;
	    $homofail = 1 if $seq =~ /c{$minrun}/i;
	    $homofail = 1 if $seq =~ /g{$minrun}/i;

	    if ($homofail) {
		$forbidden->{$trans} = 1;
		print STDERR "Filtering\t$trans\thomopolymer\t$seq\n";
	    }
	    #print "consider\t$homofail\t$seq\n";

	}

	foreach my $forbiddenFn (@$forbiddenFns) {

	    my $forbiddenFh = FileHandle->new($forbiddenFn) or die("$!:$forbiddenFn");
    
	    my $MaxTagSize = 20;
	    while (my $line = <$forbiddenFh>) {
		chomp $line;
		my ($seq) = $line =~ /^(\S+)/;
		
		# Once I was given sequences longer than 20bp, with a fixed header,
		#this removed that
		if (length($seq) > $MaxTagSize) {
		    #print length($line), "$line\n";
		    $seq = substr($seq, length($seq) - $MaxTagSize);
		}
		my $trans = $translate->{$seq};
		if (!$trans) {
		    print STDERR "Filtering\t$seq\trequested by $forbiddenFn but impossible due to lack of translation\n";
		    next;
		}
		#print "got $line $trans\n";
		$forbidden->{$trans} = 1;
		print STDERR "Filtering\t$trans\tas forbidden by\t$forbiddenFn\n";
	    }
	}
    }



    local $INPUT_RECORD_SEPARATOR = "-1\n";

    while (<STDIN>) {
	my ($top, $body) = split(/\n/, $_, 2);
	my ($bac1, $count, $bac2) = $top =~ /^(\S+)\s+(\d+)\s+(\S+)/;

	my $skipbac = 0;
	foreach my $kickoutbac (@$badbacs) {
	    if ($bac1 =~ /^${kickoutbac}_/ ) {
		print STDERR "Filtering\t$bac1\tas forbidden bac\n";
		$skipbac += 1;
		last;
	    }
	}
    
	next if $skipbac;


	my @ids = $body =~ /(-?[\d]+)/g;
	my $last = pop @ids;
	if ($last != "-1") {
	    Carp::confess();
	}

	my $newcount = 0;
	my @keep = ();
	foreach my $id (@ids) {
	    if (! $forbidden->{$id}) {
		push @keep, $id;
		$newcount++;
	    }
	}

	if (!$bac1) {
	    die("maybe you need dos2unix on the band file?");
	}

	if ($newcount) {
	    print "$bac1 $newcount $bac2\n";
	    foreach my $id (@keep) {
		print "$id\n";
	    }
	    print "-1\n";
	}
    }
}

# split a file into $size pieces. 
# $chunk can be used to take something other than the 1st piece.
sub filterN {
    my ($size, $chunk) = @_;
    $chunk = 0 unless defined $chunk;

    my $pid;
    return if $pid = open(STDOUT, "|-");
    die "cannot fork: $!" unless defined $pid;



    while (<STDIN>) {

	my ($top, $body) = split(/\n/, $_, 2);
	my ($bac1, $count, $bac2) = $top =~ /^(\S+)\s+(\d+)\s+(\S+)/;
	my @ids = $body =~ /(-?[\d]+)/g;
	my $last = pop @ids;
	if ($last != "-1") {
	    Carp::confess();
	}

	my $newcount = 0;
	my @keep = ();
	foreach my $id (@ids) {
	    my $remainder = $id % $size;
	    if ($remainder == $chunk) {
		#my $newid = int($id / $size);
		my $newid = $id;
		push @keep, $newid;
		$newcount++;
	    }
	}

	if ($newcount) {
	    print "$bac1 $newcount $bac2\n";
	    foreach my $id (@keep) {
		print "$id\n";
	    }
	    print "-1\n";
	}
    }
    exit;

}











#
# Adds 10 million to test handling of large tag #s
sub filterJan {
    my ($size, $chunk) = @_;
    $chunk = 0 unless defined $chunk;

    $INPUT_RECORD_SEPARATOR = "-1\n";


    my $pid;
    return if $pid = open(STDOUT, "|-");
    die "cannot fork: $!" unless defined $pid;



    while (<STDIN>) {

	my ($top, $body) = split(/\n/, $_, 2);
	my ($bac1, $count, $bac2) = $top =~ /^(\S+)\s+(\d+)\s+(\S+)/;
	my @ids = $body =~ /(-?[\d]+)/g;
	my $last = pop @ids;
	if ($last != "-1") {
	    Carp::confess();
	}

	my $newcount = 0;
	my @keep = ();
	foreach my $id (@ids) {
	    my $remainder = $id % $size;
	    if ($remainder == $chunk) {
		#my $newid = int($id / $size);
		#my $newid = $id + 10_000_000;
		my $newid = $id;
		#next if $id > 10000;
		push @keep, $newid;
		$newcount++;
	    }
	}

	if ($newcount) {
	    print "$bac1 $newcount $bac2\n";
	    foreach my $id (@keep) {
		print "$id\n";
	    }
	    print "-1\n";
	}
    }
    exit;

}












# Given a chosen BAC name, fine every tag, and every BAC which might
# possibly be related to this
sub BACExtract {
    my ($bacs) = @_;

    my $pid;
    return if $pid = open(STDOUT, "|-");
    die "cannot fork: $!" unless defined $pid;


    my @bacorder = ();


    my $bac2thing = {};
    my $bac2tag = {};
    my $tag2bac = {};


    while (my $rec = <STDIN>) {
	my ($top, $body) = split(/\n/, $rec, 2);
	my ($bac1, $count, $aname) = $top =~ /(\S+)\s+(\d+)\s+(\S+)/;
	push @bacorder, $bac1;

	my @tags = $body =~ /(-?[\d]+)/g;
	if (pop @tags != "-1") {
	    Carp::confess();
	}
	$bac2thing->{$bac1} = $aname;
	$bac2tag->{$bac1} = \@tags;

	foreach my $tag (@tags) {
	    push @{$tag2bac->{$tag}}, $bac1;
	}
    }



    my $wantedbacs = {};
    my $wantedtags = {};

    my @wantedbacs = ();
    foreach my $bac (@$bacs) {
	$wantedbacs->{$bac} = 0;
	push @wantedbacs, $bac;
    }

    
    while (@wantedbacs) {
	my $bac = shift @wantedbacs;
	my @wantedtags = @{$bac2tag->{$bac}};
	#print $#wantedbacs," bacs to go\n";

#	print "acquiring $bac\n";
	$wantedbacs->{$bac}= 1;

	foreach my $tag (@wantedtags) {
	    $wantedtags->{$tag} = 1;
	    my @inbacs = @{$tag2bac->{$tag}};
	    foreach my $bac2 (@inbacs) {
		
		if (! exists $wantedbacs->{$bac2}) {
		    $wantedbacs->{$bac2} = 0;
		    push @wantedbacs, $bac2;
#		    print "  also wanting bac $bac2\n";
		} else {
		    if ($wantedbacs->{$bac2}) {
#			print "  already got $bac2 ",$wantedbacs->{$bac2},"\n";
		    } else {
#			print "  not (yet) interest in $bac2 ",$wantedbacs->{$bac2},"\n";
		    }

		}
	    }
	}
    }

    foreach my $bac (@bacorder) {

	my $thing = $bac2thing->{$bac};
	next unless $thing;
	my @tags = @{$bac2tag->{$bac}};
	my $count = $#tags+1;
	push @tags, "-1";

	print "$bac $count $thing\n",join("\n",@tags),"\n";
	
    }

    exit;
}



# move all of the tags to the lowest possible #s, produce a second
# file on stderr which shows the mapping from old to new

sub compactTags {

    my $pid;
    return if $pid = open(STDOUT, "|-");
    die "cannot fork: $!" unless defined $pid;


    my $seen = {};
    my $mintag = 1;
    my $nextid = $mintag;

    my $numBACsIn = 0;
    my $numBACsOut = 0;
    
    while (my $rec = <STDIN>) {

	$numBACsIn += 1;
	


	my ($top, $body) = split(/\n/, $rec, 2);
	my ($bac1, $count, $bac2) = $top =~ /(\S+)\s+(\d+)\s+(\S+)/;
	my @ids = $body =~ /(-?[\d]+)/g;
	my $last = pop @ids;
	if ($last != "-1") {
	    Carp::confess();
	  }

	my $newcount = 0;
	my @keep = ();
	foreach my $id (@ids) {

	    my $newid;
	    if ($seen->{$id}) {
		$newid = $seen->{$id};
	    } else {
		$newid = $seen->{$id} = $nextid;
		$nextid++;
	    }


	    if ($nextid < $maxtagid) {
		$newcount += 1;
		push @keep, $newid;

	    }
	}

	# put them in the order needed
	@keep = sort {$a <=> $b} @keep;

	# print out the new block
	if ($newcount) {
	    my $out = "$bac1 $newcount $bac2\n";
	    foreach my $id (@keep) {
		$out .= "$id\n";
	    }
	    $out .= "-1\n";
	    $numBACsOut += 1;
	    print $out;
	}
    }

    #display some statistics on STDERR, with a leading # to enable filtering
    my $numtags = $nextid - $mintag;
    my $dropped = $maxtagid - $nextid;
    print STDERR "#Dropped $numBACsIn BACs down to $numBACsOut\n";
    #print STDERR "#From $numtags to $nextid ($dropped tags)\n";
    #print STDERR "#Dropped $dropped tags\n";


    # display the map from old-> new on STDERR
    foreach my $k (keys %$seen) {
	my $v = $seen->{$k};
	print STDERR "$k\t$v\n";
    }

    exit;
}












use Getopt::Long;
my $verbose;
my @kickouts = ();
my @badbacs;
my $bandsfn = "";
my $tagfn = "";
my $result = GetOptions (
			 "verbose"  => \$verbose,
			 "kickout=s" => \@kickouts,
			 "bands=s" => \$bandsfn,
			 "tag=s" => \$tagfn,
			 "exclude-bac=s" => \@badbacs,
			 );  # flag






my $fh = FileHandle->new($bandsfn) or die("$!:$bandsfn");
tagFilter(\@kickouts,$tagfn, \@badbacs);

#compactTags();
#filterN($size, $chunk);
#filterJan(1, 0);
#BACExtract(['E00BAC5236']);

while (my $rec = <$fh>) {
    print $rec;
}


