#!/usr/bin/env perl

use strict;
use warnings;
use FileHandle;
use English;
use Data::Dumper;
use Carp;
use Getopt::Long;

sub histos {
    my ($infn, $tagfn,
	$maxTagsPerBac,
	$maxBacsPerTag,
	$histfn, 
	$kickfn,
	) = @_;


    my $histofh = FileHandle->new(">".$histfn) or die("$!:histfn:$histfn");
    my $kickfh = FileHandle->new(">".$kickfn) or die("$!:kickfn:$kickfn");


    my $tag2seq = {};
    if (-f $tagfn) {

	local $INPUT_RECORD_SEPARATOR = "\n";

	my $intagfh = FileHandle->new($tagfn) or die($!);
	while (<$intagfh>) {
	    my ($tagseq, $tagid, $bacnames) = /^(\S+)\s+(\d+)\s+(.*)/;
	    next unless $bacnames;
	    my @bacs = split(/,/,$bacnames);
	    $tag2seq->{$tagid} = $tagseq;
	}
    }

    my $infh = FileHandle->new($infn) or die($!);

    local $INPUT_RECORD_SEPARATOR = "-1\n";

    my $tag2bac = {};
    my $baccount = [];
    my $bacnames = [];
    my $tagnames = [];
    while (<$infh>) {
	my ($top, $body) = split(/\n/, $_, 2);
	my ($bac1, $count, $thing) = $top =~ /^(\S+)\s+(\d+)\s+(\S+)/;
	my @ids = $body =~ /(-?[\d]+)/g;
	my $last = pop @ids;
	if ($last != "-1") {
	    Carp::confess();
	}
	#print "$bac1\t$count\n";
	push @{$bacnames->[$count]}, $bac1;
	$baccount->[$count] += 1;

	foreach my $id (@ids) {
	    push @{$tag2bac->{$id}}, $bac1;
	}
    }
    


    my @tagcount = ();
    my $maxtagcount = 0;
    foreach my $tag (keys %$tag2bac) {
	my $count = scalar(@{$tag2bac->{$tag}});
	push @{$tagnames->[$count]}, $tag;
	$tagcount[$count] += 1;
	$maxtagcount = $count if $count > $maxtagcount;
	#print "$tag\t$count\n";
    }









    print $histofh "Bac histo: How many bacs have N tags\n";
    my $maxbaccount = scalar @$baccount;
    for (my $i= 0; $i < $maxbaccount;  $i++) {
	print $histofh "$i\t",($baccount->[$i]||' ');

	#if (($maxBacsPerTag != -1) && $i > $maxBacsPerTag && $bacnames->[$i]) {
	#    print "\t";
	#    print join(", ",@{$bacnames->[$i]});
	#}


	print $histofh "\n";
    }

    if ($maxBacsPerTag != -1) {
	my @todrop = ();
	for (my $i= $maxBacsPerTag+1; $i < $maxbaccount;  $i++) {
	    next unless $bacnames->[$i];
	    my @group = @{$bacnames->[$i]};
	    #print "$i add ",$#group+1,"\n";
	    foreach my $t (@group) {
		push @todrop, $t;
		#print "    $i added $t\n";
	    }
	}

	#print Dumper \@todrop;
	
	foreach my $bac (sort @todrop) {
	    print $histofh "DropBac\t$bac\n";
	}


    }



    my $verbose = 0;

    print $histofh "Tag histo: How many tags are in N bacs\n";
    for (my $i= 0; $i <= $maxtagcount;  $i++) {
	
	print $histofh "$i\t",($tagcount[$i]||' ');


	if ($verbose && ($maxTagsPerBac != -1) && $i > $maxTagsPerBac && $tagnames->[$i]) {
	    print $histofh "\t";
	    print $histofh join(", ",@{$tagnames->[$i]});
	    print $histofh " ";
	    foreach my $tagid (@{$tagnames->[$i]}) {
		print $histofh $tag2seq->{$tagid}," ";
	    }
	    
	}
	print $histofh "\n";

    }

    if ($maxTagsPerBac != -1) {
	for (my $i= $maxTagsPerBac+1; $i <= $maxtagcount;  $i++) {
	    #print "drop $i ",Dumper $tagnames->[$i];
	    foreach my $tagid (@{$tagnames->[$i]}) {
		my $seq = $tag2seq->{$tagid};
		print $kickfh "$seq\tDropTag\t$tagid\n";
	    }
	}
    }
    
}


my $verbose;
my $maxTagsPerBac = -1;
my $maxBacsPerTag = -1;
my ($bandfn, $tagfn);
my ($histofn, $kickfn) = ('-','-');
my $result = GetOptions (
			 "histofile=s"  => \$histofn,
			 "kickfile=s"  => \$kickfn,

			 "verbose"  => \$verbose,
			 "band=s" => \$bandfn,
			 "tag=s" => \$tagfn,
			 "maxBacsPerTag=i" => \$maxBacsPerTag,
			 "maxTagsPerBac=i" => \$maxTagsPerBac,
			 ); 


#my $maxTagsPerBac = 32;
#my $maxBacsPerTag = 184;



histos($bandfn, $tagfn, $maxTagsPerBac, $maxBacsPerTag, $histofn, $kickfn);




