#!/usr/bin/env perl


use strict;
use warnings;
use FileHandle;
use English;
use Data::Dumper;
use Carp;
use Getopt::Long;

my $verbose;
my @kickouts = ();
my @mapfn = ();
my ($in, $out);
my $result = GetOptions (
			 "rename=s" => \@mapfn,
			 "in=s" => \$in,
			 "out=s" => \$out,
			 ); 




my $map = {};
foreach my $fn (@mapfn) {
    my $fh = FileHandle->new("$fn") or die("$!:$fn");
    foreach (<$fh>) {
	my ($from, $to) = split;
	$map->{$from} = $to;

	if ($from =~ /^\d/) {
	    $map->{'SP'.$from} = $to;
	}
    }
}

if (-f "$in.bands") {
    my $bandsifh = FileHandle->new("$in.bands") or die($!);
    my $bandsofh = FileHandle->new(">$out.bands") or die($!);


    {
	local $INPUT_RECORD_SEPARATOR = "-1\n";
	while (my $rec = <$bandsifh>) {
	    
	    my ($top, $body) = split(/\n/, $rec, 2);
	    my ($bac1, $count, $bac2) = $top =~ /(\S+)\s+(\d+)\s+(\S+)/;
	    my @ids = $body =~ /(-?[\d]+)/g;
	    my $last = pop @ids;
	    if ($last != "-1") {
		Carp::confess();
	      }

	    my $outbac = rebac($map, "$bac1,");
	    chomp $outbac;

	    print $bandsofh "$outbac $count $bac2\n";
	    foreach my $tagid (@ids) {
		print $bandsofh "$tagid\n";
	    }
	    print $bandsofh "-1\n";
	}
    }
}


if (-f "$in.fna") {
    my $fnaifh = FileHandle->new("$in.fna") or die($!);
    my $fnaofh = FileHandle->new(">$out.fna") or die($!);


    {
	local $INPUT_RECORD_SEPARATOR = "\n";
	while (my $line = <$fnaifh>) {
	    if ($line !~ /^>/) {
		print $fnaofh $line;
 	    } else {
		my ($start, $body) = $line =~ /^(\S+)\s+(\S+)/;
		print $fnaofh "$start ",rebac($map, $body),",\n";
	    }
	}
    }
}




if (-f "$in.tag") {
    my $tagifh = FileHandle->new("$in.tag") or die($!);
    my $tagofh = FileHandle->new(">$out.tag") or die($!);

    {
	local $INPUT_RECORD_SEPARATOR = "\n";
	while (my $line = <$tagifh>) {
	    my ($start, $body) = $line =~ /^(\S+\s+\d+)\s+(\S+),$/;
	    if ($body) {
		print $tagofh "$start ",rebac($map, $body),",\n";
	    } else {
		print $tagofh $line;
	    }
	}
    }
}



sub rebac {
    my ($map, $body) = @_;

    Carp::cluck() unless $body;

    my @newbacs = ();
    eval {
	my @oldbacs = split(/,/,$body);
	foreach my $bac (@oldbacs) {
	    
	    my ($bacbase, $platepos) = $bac =~ /^(.*?)_(.*)$/;
	    if (!$bacbase) {
		Carp::cluck("mapping for $bac prob failed. expected X_Y");
	      } else {		  
		  
		  my $outbac = $bac;
		  $outbac = $map->{$bacbase}."_$platepos" if $map->{$bacbase};
		  if (length $outbac > 16) {
		      Carp::croak("New bacname is too long '$outbac'");
		  }
		  push @newbacs, $outbac;
		  
	      }
	}
    };
    
    if ($@) {
	print "$@\n\n\n";
	exit;
    }

    return join(",",@newbacs);
}


